@@ -792,7 +792,7 @@ static void stream_free(struct avdtp_stream *stream)
 		g_source_remove(stream->timer);
 
 	if (stream->io)
-		g_io_channel_unref(stream->io);
+		close_stream(stream);
 
 	if (stream->io_id)
 		g_source_remove(stream->io_id);
@@ -1138,6 +1138,8 @@ void avdtp_unref(struct avdtp *session)
 			g_io_channel_shutdown(session->io, TRUE, NULL);
 			g_io_channel_unref(session->io);
 			session->io = NULL;
+			avdtp_set_state(session,
+					AVDTP_SESSION_STATE_DISCONNECTED);
 		}
 
 		if (session->io)
@@ -2434,7 +2436,7 @@ drop:
 	g_io_channel_shutdown(chan, TRUE, NULL);
 }
 
-static int l2cap_connect(struct avdtp *session)
+static GIOChannel *l2cap_connect(struct avdtp *session)
 {
 	GError *err = NULL;
 	GIOChannel *io;
@@ -2448,12 +2450,10 @@ static int l2cap_connect(struct avdtp *session)
 	if (!io) {
 		error("%s", err->message);
 		g_error_free(err);
-		return -EIO;
+		return NULL;
 	}
 
-	g_io_channel_unref(io);
-
-	return 0;
+	return io;
 }
 
 static void queue_request(struct avdtp *session, struct pending_req *req,
@@ -2588,8 +2588,8 @@ static int send_req(struct avdtp *session, gboolean priority,
 	int err;
 
 	if (session->state == AVDTP_SESSION_STATE_DISCONNECTED) {
-		err = l2cap_connect(session);
-		if (err < 0)
+		session->io = l2cap_connect(session);
+		if (!session->io)
 			goto failed;
 		avdtp_set_state(session, AVDTP_SESSION_STATE_CONNECTING);
 	}
@@ -2632,8 +2632,10 @@ static int send_request(struct avdtp *session, gboolean priority,
 {
 	struct pending_req *req;
 
-	if (stream && stream->abort_int && signal_id != AVDTP_ABORT)
+	if (stream && stream->abort_int && signal_id != AVDTP_ABORT) {
+		DBG("Unable to send requests while aborting");
 		return -EINVAL;
+	}
 
 	req = g_new0(struct pending_req, 1);
 	req->signal_id = signal_id;
@@ -2765,7 +2767,8 @@ static gboolean avdtp_open_resp(struct avdtp *session, struct avdtp_stream *stre
 {
 	struct avdtp_local_sep *sep = stream->lsep;
 
-	if (l2cap_connect(session) < 0) {
+	stream->io = l2cap_connect(session);
+	if (!stream->io) {
 		avdtp_sep_set_state(session, sep, AVDTP_STATE_IDLE);
 		return FALSE;
 	}
@@ -3546,8 +3549,7 @@ int avdtp_abort(struct avdtp *session, struct avdtp_stream *stream)
 	if (!g_slist_find(session->streams, stream))
 		return -EINVAL;
 
-	if (stream->lsep->state == AVDTP_STATE_IDLE ||
-			stream->lsep->state == AVDTP_STATE_ABORTING)
+	if (stream->lsep->state == AVDTP_STATE_ABORTING)
 		return -EINVAL;
 
 	if (session->req && stream == session->req->stream)
