@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -93,7 +93,7 @@ int bnep_init(void)
 	}
 
 	/* Temporary ioctl compatibility hack */
-	{ 
+	{
 		struct bnep_connlist_req req;
 		struct bnep_conninfo ci[1];
 
@@ -128,7 +128,7 @@ int bnep_show_connections(void)
 {
 	struct bnep_connlist_req req;
 	struct bnep_conninfo ci[48];
-	int i;
+	unsigned int i;
 
 	req.cnum = 48;
 	req.ci   = ci;
@@ -137,7 +137,7 @@ int bnep_show_connections(void)
 		return -1;
 	}
 
-	for (i=0; i < req.cnum; i++) {
+	for (i = 0; i < req.cnum; i++) {
 		printf("%s %s %s\n", ci[i].device,
 			batostr((bdaddr_t *) ci[i].dst),
 			bnep_svc2str(ci[i].role));
@@ -162,7 +162,7 @@ int bnep_kill_all_connections(void)
 {
 	struct bnep_connlist_req req;
 	struct bnep_conninfo ci[48];
-	int i;
+	unsigned int i;
 
 	req.cnum = 48;
 	req.ci   = ci;
@@ -171,7 +171,7 @@ int bnep_kill_all_connections(void)
 		return -1;
 	}
 
-	for (i=0; i < req.cnum; i++) {
+	for (i = 0; i < req.cnum; i++) {
 		struct bnep_conndel_req req;
 		memcpy(req.dst, ci[i].dst, ETH_ALEN);
 		req.flags = 0;
@@ -194,19 +194,19 @@ static int bnep_connadd(int sk, uint16_t role, char *dev)
 	return 0;
 }
 
-struct __service_16 { 
+struct __service_16 {
 	uint16_t dst;
 	uint16_t src;
 } __attribute__ ((packed));
 
-struct __service_32 { 
+struct __service_32 {
 	uint16_t unused1;
 	uint16_t dst;
 	uint16_t unused2;
 	uint16_t src;
 } __attribute__ ((packed));
 
-struct __service_128 { 
+struct __service_128 {
 	uint16_t unused1;
 	uint16_t dst;
 	uint16_t unused2[8];
@@ -219,7 +219,7 @@ int bnep_accept_connection(int sk, uint16_t role, char *dev)
 	struct bnep_setup_conn_req *req;
 	struct bnep_control_rsp *rsp;
 	unsigned char pkt[BNEP_MTU];
-	int r;
+	ssize_t r;
 
 	r = recv(sk, pkt, BNEP_MTU, 0);
 	if (r <= 0)
@@ -227,7 +227,7 @@ int bnep_accept_connection(int sk, uint16_t role, char *dev)
 
 	errno = EPROTO;
 
-	if (r < sizeof(*req))
+	if ((size_t) r < sizeof(*req))
 		return -1;
 
 	req = (void *) pkt;
@@ -246,7 +246,7 @@ int bnep_accept_connection(int sk, uint16_t role, char *dev)
 	return bnep_connadd(sk, role, dev);
 }
 
-/* Create BNEP connection 
+/* Create BNEP connection
  * sk      - Connect L2CAP socket
  * role    - Local role
  * service - Remote service
@@ -259,7 +259,7 @@ int bnep_create_connection(int sk, uint16_t role, uint16_t svc, char *dev)
 	struct __service_16 *s;
 	struct timeval timeo;
 	unsigned char pkt[BNEP_MTU];
-	int r;
+	ssize_t r;
 
 	/* Send request */
 	req = (void *) pkt;
@@ -292,9 +292,9 @@ receive:
 
 	errno = EPROTO;
 
-	if (r < sizeof(*rsp))
+	if ((size_t) r < sizeof(*rsp))
 		return -1;
-	
+
 	rsp = (void *) pkt;
 	if (rsp->type != BNEP_CONTROL)
 		return -1;
