@@ -263,14 +263,20 @@ static gint cmp_mdl_state(gconstpointer a, gconstpointer b)
 		return 1;
 }
 
-static void free_mcl_priv_data(struct mcap_mcl *mcl)
+static void free_mcap_mdl_op(struct mcap_mdl_op_cb *op)
 {
-	struct mcap_mdl_op_cb *op = mcl->priv_data;
-
 	if (op->destroy)
 		op->destroy(op->user_data);
 
-	g_free(mcl->priv_data);
+	if (op->mdl)
+		mcap_mdl_unref(op->mdl);
+
+	g_free(op);
+}
+
+static void free_mcl_priv_data(struct mcap_mcl *mcl)
+{
+	free_mcap_mdl_op(mcl->priv_data);
 	mcl->priv_data = NULL;
 }
 
@@ -290,7 +296,7 @@ static void mcap_notify_error(struct mcap_mcl *mcl, GError *err)
 		l = g_slist_find_custom(mcl->mdls, &st, cmp_mdl_state);
 		mdl = l->data;
 		mcl->mdls = g_slist_remove(mcl->mdls, mdl);
-		free_mdl(mdl);
+		mcap_mdl_unref(mdl);
 		update_mcl_state(mcl);
 		con->cb.op_conf(NULL, 0, err, con->user_data);
 		break;
@@ -451,7 +457,7 @@ static gboolean wait_response_timer(gpointer data)
 	mcap_notify_error(mcl, gerr);
 
 	g_error_free(gerr);
-	mcl->ms->mcl_disconnected_cb(mcl, mcl->ms->user_data);
+	mcl->mi->mcl_disconnected_cb(mcl, mcl->mi->user_data);
 	mcap_cache_mcl(mcl);
 
 	return FALSE;
@@ -484,7 +490,7 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 	mdl->state = MDL_WAITING;
 
 	con = g_new0(struct mcap_mdl_op_cb, 1);
-	con->mdl = mdl;
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.op_conf = connect_cb;
 	con->destroy = destroy;
 	con->user_data = user_data;
@@ -492,7 +498,7 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 	cmd = create_mdl_req(id, mdepid, conf);
 	if (!mcap_send_std_opcode(mcl, cmd, sizeof(mcap_md_create_mdl_req),
 									err)) {
-		free_mdl(mdl);
+		mcap_mdl_unref(con->mdl);
 		g_free(con);
 		g_free(cmd);
 		return FALSE;
@@ -501,7 +507,8 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 	mcl->state = MCL_ACTIVE;
 	mcl->priv_data = con;
 
-	mcl->mdls = g_slist_insert_sorted(mcl->mdls, mdl, compare_mdl);
+	mcl->mdls = g_slist_insert_sorted(mcl->mdls, mcap_mdl_ref(mdl),
+								compare_mdl);
 	mcl->tid = g_timeout_add_seconds(RESPONSE_TIMER, wait_response_timer,
 									mcl);
 	return TRUE;
@@ -522,18 +529,17 @@ gboolean mcap_reconnect_mdl(struct mcap_mdl *mdl,
 					"MDL is not closed");
 		return FALSE;
 	}
-	con = g_new0(struct mcap_mdl_op_cb, 1);
 
 	cmd = create_req(MCAP_MD_RECONNECT_MDL_REQ, mdl->mdlid);
 	if (!mcap_send_std_opcode(mcl, cmd, sizeof(mcap_md_req), err)) {
-		g_free(con);
 		g_free(cmd);
 		return FALSE;
 	}
 
 	mdl->state = MDL_WAITING;
 
-	con->mdl = mdl;
+	con = g_new0(struct mcap_mdl_op_cb, 1);
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.op = reconnect_cb;
 	con->destroy = destroy;
 	con->user_data = user_data;
@@ -630,12 +636,13 @@ gboolean mcap_delete_mdl(struct mcap_mdl *mdl, mcap_mdl_notify_cb delete_cb,
 	mdl->state = MDL_DELETING;
 
 	con = g_new0(struct mcap_mdl_op_cb, 1);
-	con->mdl = mdl;
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.notify = delete_cb;
 	con->destroy = destroy;
 	con->user_data = user_data;
 
 	if (!send_delete_req(mcl, con, mdl->mdlid, err)) {
+		mcap_mdl_unref(con->mdl);
 		g_free(con);
 		return FALSE;
 	}
@@ -658,15 +665,14 @@ gboolean mcap_mdl_abort(struct mcap_mdl *mdl, mcap_mdl_notify_cb abort_cb,
 		return FALSE;
 	}
 
-	con = g_new0(struct mcap_mdl_op_cb, 1);
 	cmd = create_req(MCAP_MD_ABORT_MDL_REQ, mdl->mdlid);
 	if (!mcap_send_std_opcode(mcl, cmd, sizeof(mcap_md_req), err)) {
-		g_free(con);
 		g_free(cmd);
 		return FALSE;
 	}
 
-	con->mdl = mdl;
+	con = g_new0(struct mcap_mdl_op_cb, 1);
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.notify = abort_cb;
 	con->destroy = destroy;
 	con->user_data = user_data;
@@ -742,7 +748,7 @@ static void close_mcl(struct mcap_mcl *mcl, gboolean cache_requested)
 	if (save)
 		return;
 
-	g_slist_foreach(mcl->mdls, (GFunc) free_mdl, NULL);
+	g_slist_foreach(mcl->mdls, (GFunc) mcap_mdl_unref, NULL);
 	g_slist_free(mcl->mdls);
 	mcl->mdls = NULL;
 }
@@ -766,10 +772,10 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 	if (mcl->ctrl & MCAP_CTRL_CACHED)
 		return;
 
-	mcl->ms->mcls = g_slist_remove(mcl->ms->mcls, mcl);
+	mcl->mi->mcls = g_slist_remove(mcl->mi->mcls, mcl);
 
 	if (mcl->ctrl & MCAP_CTRL_NOCACHE) {
-		mcl->ms->cached = g_slist_remove(mcl->ms->cached, mcl);
+		mcl->mi->cached = g_slist_remove(mcl->mi->cached, mcl);
 		mcap_mcl_release(mcl);
 		mcap_mcl_unref(mcl);
 		return;
@@ -777,12 +783,12 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 
 	DBG("Caching MCL");
 
-	len = g_slist_length(mcl->ms->cached);
+	len = g_slist_length(mcl->mi->cached);
 	if (len == MAX_CACHED) {
 		/* Remove the latest cached mcl */
-		l = g_slist_last(mcl->ms->cached);
+		l = g_slist_last(mcl->mi->cached);
 		last = l->data;
-		mcl->ms->cached = g_slist_remove(mcl->ms->cached, last);
+		mcl->mi->cached = g_slist_remove(mcl->mi->cached, last);
 		last->ctrl &= ~MCAP_CTRL_CACHED;
 		if (last->ctrl & MCAP_CTRL_CONN) {
 			/* We have to release this MCL if */
@@ -790,12 +796,12 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 			last->ctrl |= MCAP_CTRL_FREE;
 		} else {
 			mcap_mcl_release(last);
-			last->ms->mcl_uncached_cb(last, last->ms->user_data);
+			last->mi->mcl_uncached_cb(last, last->mi->user_data);
 		}
 		mcap_mcl_unref(last);
 	}
 
-	mcl->ms->cached = g_slist_prepend(mcl->ms->cached, mcl);
+	mcl->mi->cached = g_slist_prepend(mcl->mi->cached, mcl);
 	mcl->ctrl |= MCAP_CTRL_CACHED;
 	mcap_mcl_shutdown(mcl);
 }
@@ -807,8 +813,8 @@ static void mcap_uncache_mcl(struct mcap_mcl *mcl)
 
 	DBG("Got MCL from cache");
 
-	mcl->ms->cached = g_slist_remove(mcl->ms->cached, mcl);
-	mcl->ms->mcls = g_slist_prepend(mcl->ms->mcls, mcl);
+	mcl->mi->cached = g_slist_remove(mcl->mi->cached, mcl);
+	mcl->mi->mcls = g_slist_prepend(mcl->mi->mcls, mcl);
 	mcl->ctrl &= ~MCAP_CTRL_CACHED;
 	mcl->ctrl &= ~MCAP_CTRL_FREE;
 }
@@ -834,7 +840,7 @@ void mcap_close_mcl(struct mcap_mcl *mcl, gboolean cache)
 	} else if ((mcl->ctrl & MCAP_CTRL_CACHED) &&
 					(mcl->ctrl & MCAP_CTRL_NOCACHE)) {
 		mcl->ctrl &= ~MCAP_CTRL_CACHED;
-		mcl->ms->cached = g_slist_remove(mcl->ms->cached, mcl);
+		mcl->mi->cached = g_slist_remove(mcl->mi->cached, mcl);
 		mcap_mcl_release(mcl);
 		mcap_mcl_unref(mcl);
 	}
@@ -859,7 +865,7 @@ void mcap_mcl_unref(struct mcap_mcl *mcl)
 		return;
 
 	mcap_mcl_release(mcl);
-
+	mcap_instance_unref(mcl->mi);
 	g_free(mcl->cb);
 	g_free(mcl);
 }
@@ -952,7 +958,7 @@ static void mcap_del_mdl(gpointer elem, gpointer user_data)
 	if (notify)
 		mdl->mcl->cb->mdl_deleted(mdl, mdl->mcl->cb->user_data);
 
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 }
 
 static gboolean check_cmd_req_length(struct mcap_mcl *mcl, void *cmd,
@@ -1041,7 +1047,8 @@ static void process_md_create_mdl_req(struct mcap_mcl *mcl, void *cmd,
 		mdl = g_new0(struct mcap_mdl, 1);
 		mdl->mcl = mcap_mcl_ref(mcl);
 		mdl->mdlid = mdl_id;
-		mcl->mdls = g_slist_insert_sorted(mcl->mdls, mdl, compare_mdl);
+		mcl->mdls = g_slist_insert_sorted(mcl->mdls, mcap_mdl_ref(mdl),
+								compare_mdl);
 	} else if (mdl->state == MDL_CONNECTED) {
 		/* MCAP specification says that we should close the MCL if
 		 * it is open when we receive a MD_CREATE_MDL_REQ */
@@ -1351,7 +1358,7 @@ static gboolean process_md_create_mdl_rsp(struct mcap_mcl *mcl,
 fail:
 	connect_cb(NULL, 0, gerr, user_data);
 	mcl->mdls = g_slist_remove(mcl->mdls, mdl);
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 	g_error_free(gerr);
 	update_mcl_state(mcl);
 	return close;
@@ -1387,7 +1394,7 @@ static gboolean process_md_reconnect_mdl_rsp(struct mcap_mcl *mcl,
 	/* Remove cached mdlid */
 	mcl->mdls = g_slist_remove(mcl->mdls, mdl);
 	mcl->cb->mdl_deleted(mdl, mcl->cb->user_data);
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 
 	return close;
 }
@@ -1413,7 +1420,8 @@ static gboolean process_md_abort_mdl_rsp(struct mcap_mcl *mcl,
 
 	if (len >= sizeof(mcap_rsp) && rsp->rc == MCAP_INVALID_MDL) {
 		mcl->mdls = g_slist_remove(mcl->mdls, mdl);
-		free_mdl(mdl);
+		mcl->cb->mdl_deleted(mdl, mcl->cb->user_data);
+		mcap_mdl_unref(mdl);
 	}
 
 	if (gerr)
@@ -1446,7 +1454,7 @@ static void check_mdl_del_err(struct mcap_mdl *mdl, mcap_rsp *rsp)
 
 	/* MDL does not exist in remote side, we can delete it */
 	mdl->mcl->mdls = g_slist_remove(mdl->mcl->mdls, mdl);
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 }
 
 static gboolean process_md_delete_mdl_rsp(struct mcap_mcl *mcl, mcap_rsp *rsp,
@@ -1496,18 +1504,14 @@ static gboolean process_md_delete_mdl_rsp(struct mcap_mcl *mcl, mcap_rsp *rsp,
 
 static void post_process_rsp(struct mcap_mcl *mcl, struct mcap_mdl_op_cb *op)
 {
-	if (op->destroy)
-		op->destroy(op->user_data);
-
 	if (mcl->priv_data != op) {
 		/* Queued MCAP request in some callback. */
-		/* We sould not delete the mcl private data */
-		g_free(op);
+		/* We should not delete the mcl private data */
+		free_mcap_mdl_op(op);
 	} else {
-		/* This is not queued requets. It's safe */
+		/* This is not a queued request. It's safe */
 		/* delete the mcl private data here. */
-		g_free(mcl->priv_data);
-		mcl->priv_data = NULL;
+		free_mcl_priv_data(mcl);
 	}
 }
 
@@ -1543,7 +1547,7 @@ static void proc_response(struct mcap_mcl *mcl, void *buf, uint32_t len)
 	}
 
 	if (close) {
-		mcl->ms->mcl_disconnected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_disconnected_cb(mcl, mcl->mi->user_data);
 		mcap_cache_mcl(mcl);
 	}
 }
@@ -1640,16 +1644,6 @@ static void mcap_connect_mdl_cb(GIOChannel *chan, GError *conn_err,
 	cb(mdl, conn_err, user_data);
 }
 
-static void mdl_io_destroy(gpointer data)
-{
-	struct mcap_mdl_op_cb *con = data;
-
-	if (con->destroy)
-		con->destroy(con->user_data);
-
-	g_free(con);
-}
-
 gboolean mcap_connect_mdl(struct mcap_mdl *mdl, uint8_t mode,
 					uint16_t dcpsm,
 					mcap_mdl_operation_cb connect_cb,
@@ -1672,25 +1666,24 @@ gboolean mcap_connect_mdl(struct mcap_mdl *mdl, uint8_t mode,
 	}
 
 	con = g_new0(struct mcap_mdl_op_cb, 1);
-	con->mdl = mdl;
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.op = connect_cb;
 	con->destroy = destroy;
 	con->user_data = user_data;
 
-	/* TODO: Check if BtIOType is ERTM or Streaming before continue */
-
 	mdl->dc = bt_io_connect(BT_IO_L2CAP, mcap_connect_mdl_cb, con,
-				mdl_io_destroy, err,
-				BT_IO_OPT_SOURCE_BDADDR, &mdl->mcl->ms->src,
+				(GDestroyNotify) free_mcap_mdl_op, err,
+				BT_IO_OPT_SOURCE_BDADDR, &mdl->mcl->mi->src,
 				BT_IO_OPT_DEST_BDADDR, &mdl->mcl->addr,
 				BT_IO_OPT_PSM, dcpsm,
 				BT_IO_OPT_MTU, MCAP_DC_MTU,
-				BT_IO_OPT_SEC_LEVEL, mdl->mcl->ms->sec,
+				BT_IO_OPT_SEC_LEVEL, mdl->mcl->mi->sec,
 				BT_IO_OPT_MODE, mode,
 				BT_IO_OPT_INVALID);
 	if (!mdl->dc) {
 		DBG("MDL Connection error");
 		mdl->state = MDL_CLOSED;
+		mcap_mdl_unref(con->mdl);
 		g_free(con);
 		return FALSE;
 	}
@@ -1726,7 +1719,7 @@ fail:
 			mcap_notify_error(mcl, gerr);
 			g_error_free(gerr);
 		}
-		mcl->ms->mcl_disconnected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_disconnected_cb(mcl, mcl->mi->user_data);
 	}
 	mcap_cache_mcl(mcl);
 	return FALSE;
@@ -1747,7 +1740,7 @@ static void mcap_connect_mcl_cb(GIOChannel *chan, GError *conn_err,
 	if (conn_err) {
 		if (mcl->ctrl & MCAP_CTRL_FREE) {
 			mcap_mcl_release(mcl);
-			mcl->ms->mcl_uncached_cb(mcl, mcl->ms->user_data);
+			mcl->mi->mcl_uncached_cb(mcl, mcl->mi->user_data);
 		}
 		connect_cb(NULL, conn_err, data);
 		return;
@@ -1755,7 +1748,7 @@ static void mcap_connect_mcl_cb(GIOChannel *chan, GError *conn_err,
 
 	ba2str(&mcl->addr, dstaddr);
 
-	aux = find_mcl(mcl->ms->mcls, &mcl->addr);
+	aux = find_mcl(mcl->mi->mcls, &mcl->addr);
 	if (aux) {
 		/* Double MCL connection case */
 		error("MCL error: Device %s is already connected", dstaddr);
@@ -1777,7 +1770,7 @@ static void mcap_connect_mcl_cb(GIOChannel *chan, GError *conn_err,
 		mcap_uncache_mcl(mcl);
 	else {
 		mcl->ctrl &= ~MCAP_CTRL_FREE;
-		mcl->ms->mcls = g_slist_prepend(mcl->ms->mcls,
+		mcl->mi->mcls = g_slist_prepend(mcl->mi->mcls,
 							mcap_mcl_ref(mcl));
 	}
 
@@ -1810,7 +1803,7 @@ static void mcl_io_destroy(gpointer data)
 	g_free(con);
 }
 
-gboolean mcap_create_mcl(struct mcap_instance *ms,
+gboolean mcap_create_mcl(struct mcap_instance *mi,
 				const bdaddr_t *addr,
 				uint16_t ccpsm,
 				mcap_mcl_connect_cb connect_cb,
@@ -1821,17 +1814,17 @@ gboolean mcap_create_mcl(struct mcap_instance *ms,
 	struct mcap_mcl *mcl;
 	struct connect_mcl *con;
 
-	mcl = find_mcl(ms->mcls, addr);
+	mcl = find_mcl(mi->mcls, addr);
 	if (mcl) {
 		g_set_error(err, MCAP_ERROR, MCAP_ERROR_ALREADY_EXISTS,
 					"MCL is already connected.");
 		return FALSE;
 	}
 
-	mcl = find_mcl(ms->cached, addr);
+	mcl = find_mcl(mi->cached, addr);
 	if (!mcl) {
 		mcl = g_new0(struct mcap_mcl, 1);
-		mcl->ms = ms;
+		mcl->mi = mcap_instance_ref(mi);
 		mcl->state = MCL_IDLE;
 		bacpy(&mcl->addr, addr);
 		set_default_cb(mcl);
@@ -1848,18 +1841,18 @@ gboolean mcap_create_mcl(struct mcap_instance *ms,
 
 	mcl->cc = bt_io_connect(BT_IO_L2CAP, mcap_connect_mcl_cb, con,
 				mcl_io_destroy, err,
-				BT_IO_OPT_SOURCE_BDADDR, &ms->src,
+				BT_IO_OPT_SOURCE_BDADDR, &mi->src,
 				BT_IO_OPT_DEST_BDADDR, addr,
 				BT_IO_OPT_PSM, ccpsm,
 				BT_IO_OPT_MTU, MCAP_CC_MTU,
-				BT_IO_OPT_SEC_LEVEL, ms->sec,
+				BT_IO_OPT_SEC_LEVEL, mi->sec,
 				BT_IO_OPT_MODE, L2CAP_MODE_ERTM,
 				BT_IO_OPT_INVALID);
 	if (!mcl->cc) {
 		mcl->ctrl &= ~MCAP_CTRL_CONN;
 		if (mcl->ctrl & MCAP_CTRL_FREE) {
 			mcap_mcl_release(mcl);
-			mcl->ms->mcl_uncached_cb(mcl, mcl->ms->user_data);
+			mcl->mi->mcl_uncached_cb(mcl, mcl->mi->user_data);
 		}
 		mcap_mcl_unref(con->mcl);
 		g_free(con);
@@ -1872,7 +1865,7 @@ gboolean mcap_create_mcl(struct mcap_instance *ms,
 static void connect_dc_event_cb(GIOChannel *chan, GError *gerr,
 							gpointer user_data)
 {
-	struct mcap_instance *ms = user_data;
+	struct mcap_instance *mi = user_data;
 	struct mcap_mcl *mcl;
 	struct mcap_mdl *mdl;
 	GError *err = NULL;
@@ -1891,7 +1884,7 @@ static void connect_dc_event_cb(GIOChannel *chan, GError *gerr,
 		goto drop;
 	}
 
-	mcl = find_mcl(ms->mcls, &dst);
+	mcl = find_mcl(mi->mcls, &dst);
 	if (!mcl || mcl->state != MCL_PENDING)
 		goto drop;
 
@@ -1923,7 +1916,7 @@ static void set_mcl_conf(GIOChannel *chan, struct mcap_mcl *mcl)
 	if (reconn)
 		mcap_uncache_mcl(mcl);
 	else
-		mcl->ms->mcls = g_slist_prepend(mcl->ms->mcls,
+		mcl->mi->mcls = g_slist_prepend(mcl->mi->mcls,
 							mcap_mcl_ref(mcl));
 
 	mcl->wid = g_io_add_watch(mcl->cc,
@@ -1932,15 +1925,15 @@ static void set_mcl_conf(GIOChannel *chan, struct mcap_mcl *mcl)
 
 	/* Callback to report new MCL */
 	if (reconn)
-		mcl->ms->mcl_reconnected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_reconnected_cb(mcl, mcl->mi->user_data);
 	else
-		mcl->ms->mcl_connected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_connected_cb(mcl, mcl->mi->user_data);
 }
 
 static void connect_mcl_event_cb(GIOChannel *chan, GError *gerr,
 							gpointer user_data)
 {
-	struct mcap_instance *ms = user_data;
+	struct mcap_instance *mi = user_data;
 	struct mcap_mcl *mcl;
 	bdaddr_t dst;
 	char address[18], srcstr[18];
@@ -1959,18 +1952,18 @@ static void connect_mcl_event_cb(GIOChannel *chan, GError *gerr,
 		goto drop;
 	}
 
-	ba2str(&ms->src, srcstr);
-	mcl = find_mcl(ms->mcls, &dst);
+	ba2str(&mi->src, srcstr);
+	mcl = find_mcl(mi->mcls, &dst);
 	if (mcl) {
 		error("Control channel already created with %s on adapter %s",
 				address, srcstr);
 		goto drop;
 	}
 
-	mcl = find_mcl(ms->cached, &dst);
+	mcl = find_mcl(mi->cached, &dst);
 	if (!mcl) {
 		mcl = g_new0(struct mcap_mcl, 1);
-		mcl->ms = ms;
+		mcl->mi = mcap_instance_ref(mi);
 		bacpy(&mcl->addr, &dst);
 		set_default_cb(mcl);
 		mcl->next_mdl = (rand() % MCAP_MDLID_FINAL) + 1;
@@ -1995,7 +1988,7 @@ struct mcap_instance *mcap_create_instance(bdaddr_t *src,
 					gpointer user_data,
 					GError **gerr)
 {
-	struct mcap_instance *ms;
+	struct mcap_instance *mi;
 
 	if (sec < BT_IO_SEC_MEDIUM) {
 		g_set_error(gerr, MCAP_ERROR, MCAP_ERROR_INVALID_ARGS,
@@ -2011,55 +2004,55 @@ struct mcap_instance *mcap_create_instance(bdaddr_t *src,
 		return NULL;
 	}
 
-	ms = g_new0(struct mcap_instance, 1);
+	mi = g_new0(struct mcap_instance, 1);
 
-	bacpy(&ms->src, src);
+	bacpy(&mi->src, src);
 
-	ms->sec = sec;
-	ms->mcl_connected_cb = mcl_connected;
-	ms->mcl_reconnected_cb = mcl_reconnected;
-	ms->mcl_disconnected_cb = mcl_disconnected;
-	ms->mcl_uncached_cb = mcl_uncached;
-	ms->mcl_sync_infoind_cb = mcl_sync_info_ind;
-	ms->user_data = user_data;
-	ms->csp_enabled = FALSE;
+	mi->sec = sec;
+	mi->mcl_connected_cb = mcl_connected;
+	mi->mcl_reconnected_cb = mcl_reconnected;
+	mi->mcl_disconnected_cb = mcl_disconnected;
+	mi->mcl_uncached_cb = mcl_uncached;
+	mi->mcl_sync_infoind_cb = mcl_sync_info_ind;
+	mi->user_data = user_data;
+	mi->csp_enabled = FALSE;
 
 	/* Listen incoming connections in control channel */
-	ms->ccio = bt_io_listen(BT_IO_L2CAP, connect_mcl_event_cb, NULL, ms,
+	mi->ccio = bt_io_listen(BT_IO_L2CAP, connect_mcl_event_cb, NULL, mi,
 				NULL, gerr,
-				BT_IO_OPT_SOURCE_BDADDR, &ms->src,
+				BT_IO_OPT_SOURCE_BDADDR, &mi->src,
 				BT_IO_OPT_PSM, ccpsm,
 				BT_IO_OPT_MTU, MCAP_CC_MTU,
 				BT_IO_OPT_SEC_LEVEL, sec,
 				BT_IO_OPT_MODE, L2CAP_MODE_ERTM,
 				BT_IO_OPT_INVALID);
-	if (!ms->ccio) {
+	if (!mi->ccio) {
 		error("%s", (*gerr)->message);
-		g_free(ms);
+		g_free(mi);
 		return NULL;
 	}
 
 	/* Listen incoming connections in data channels */
-	ms->dcio = bt_io_listen(BT_IO_L2CAP, connect_dc_event_cb, NULL, ms,
+	mi->dcio = bt_io_listen(BT_IO_L2CAP, connect_dc_event_cb, NULL, mi,
 				NULL, gerr,
-				BT_IO_OPT_SOURCE_BDADDR, &ms->src,
+				BT_IO_OPT_SOURCE_BDADDR, &mi->src,
 				BT_IO_OPT_PSM, dcpsm,
 				BT_IO_OPT_MTU, MCAP_DC_MTU,
 				BT_IO_OPT_SEC_LEVEL, sec,
 				BT_IO_OPT_INVALID);
-	if (!ms->dcio) {
-		g_io_channel_shutdown(ms->ccio, TRUE, NULL);
-		g_io_channel_unref(ms->ccio);
-		ms->ccio = NULL;
+	if (!mi->dcio) {
+		g_io_channel_shutdown(mi->ccio, TRUE, NULL);
+		g_io_channel_unref(mi->ccio);
+		mi->ccio = NULL;
 		error("%s", (*gerr)->message);
-		g_free(ms);
+		g_free(mi);
 		return NULL;
 	}
 
 	/* Initialize random seed to generate mdlids for this instance */
 	srand(time(NULL));
 
-	return ms;
+	return mcap_instance_ref(mi);;
 }
 
 void mcap_release_instance(struct mcap_instance *mi)
@@ -2085,6 +2078,7 @@ void mcap_release_instance(struct mcap_instance *mi)
 		mcap_mcl_release(l->data);
 		mcap_mcl_unref(l->data);
 	}
+
 	g_slist_free(mi->mcls);
 	mi->mcls = NULL;
 
@@ -2092,9 +2086,30 @@ void mcap_release_instance(struct mcap_instance *mi)
 		mcap_mcl_release(l->data);
 		mcap_mcl_unref(l->data);
 	}
+
 	g_slist_free(mi->cached);
 	mi->cached = NULL;
+}
 
+struct mcap_instance *mcap_instance_ref(struct mcap_instance *mi)
+{
+	mi->ref++;
+
+	DBG("mcap_instance_ref(%p): ref=%d", mi, mi->ref);
+
+	return mi;
+}
+
+void mcap_instance_unref(struct mcap_instance *mi)
+{
+	mi->ref--;
+
+	DBG("mcap_instance_unref(%p): ref=%d", mi, mi->ref);
+
+	if (mi->ref > 0)
+		return;
+
+	mcap_release_instance(mi);
 	g_free(mi);
 }
 
@@ -2133,3 +2148,37 @@ uint16_t mcap_get_data_psm(struct mcap_instance *mi, GError **err)
 
 	return lpsm;
 }
+
+gboolean mcap_set_data_chan_mode(struct mcap_instance *mi, uint8_t mode,
+								GError **err)
+{
+	if (!(mi && mi->dcio)) {
+		g_set_error(err, MCAP_ERROR, MCAP_ERROR_INVALID_ARGS,
+						"Invalid MCAP instance");
+		return FALSE;
+	}
+
+	return bt_io_set(mi->dcio, BT_IO_L2CAP, err, BT_IO_OPT_MODE, mode,
+							BT_IO_OPT_INVALID);
+}
+
+struct mcap_mdl *mcap_mdl_ref(struct mcap_mdl *mdl)
+{
+	mdl->ref++;
+
+	DBG("mcap_mdl_ref(%p): ref=%d", mdl, mdl->ref);
+
+	return mdl;
+}
+
+void mcap_mdl_unref(struct mcap_mdl *mdl)
+{
+	mdl->ref--;
+
+	DBG("mcap_mdl_unref(%p): ref=%d", mdl, mdl->ref);
+
+	if (mdl->ref > 0)
+		return;
+
+	free_mdl(mdl);
+}
