@@ -55,6 +55,9 @@ static struct controller_info {
 	uint8_t type;
 	bdaddr_t bdaddr;
 	uint8_t features[8];
+	uint16_t manufacturer;
+	uint8_t hci_ver;
+	uint16_t hci_rev;
 	gboolean enabled;
 	uint8_t mode;
 } *controllers = NULL;
@@ -68,9 +71,9 @@ static uint16_t mgmt_revision = 0;
 static void read_version_complete(int sk, void *buf, size_t len)
 {
 	struct mgmt_hdr hdr;
-	struct mgmt_read_version_rp *rp = buf;
+	struct mgmt_rp_read_version *rp = buf;
 
-	if (len < MGMT_READ_VERSION_RP_SIZE) {
+	if (len < sizeof(*rp)) {
 		error("Too small read version complete event");
 		return;
 	}
@@ -98,13 +101,15 @@ static void add_controller(uint16_t index)
 	memset(&controllers[index], 0, sizeof(struct controller_info));
 
 	controllers[index].valid = TRUE;
+
+	DBG("Added controller %u", index);
 }
 
 static void read_info(int sk, uint16_t index)
 {
-	char buf[MGMT_HDR_SIZE + MGMT_READ_INFO_CP_SIZE];
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_read_info)];
 	struct mgmt_hdr *hdr = (void *) buf;
-	struct mgmt_read_info_cp *cp = (void *) &buf[sizeof(*hdr)];
+	struct mgmt_cp_read_info *cp = (void *) &buf[sizeof(*hdr)];
 
 	memset(buf, 0, sizeof(buf));
 	hdr->opcode = MGMT_OP_READ_INFO;
@@ -117,11 +122,57 @@ static void read_info(int sk, uint16_t index)
 						strerror(errno), errno);
 }
 
+static void mgmt_index_added(int sk, void *buf, size_t len)
+{
+	struct mgmt_ev_index_added *ev = buf;
+	uint16_t index;
+
+	if (len < sizeof(*ev)) {
+		error("Too small index added event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	add_controller(index);
+	read_info(sk, index);
+}
+
+static void remove_controller(uint16_t index)
+{
+	if (index > max_index)
+		return;
+
+	if (!controllers[index].valid)
+		return;
+
+	manager_unregister_adapter(index);
+
+	memset(&controllers[index], 0, sizeof(struct controller_info));
+
+	DBG("Removed controller %u", index);
+}
+
+static void mgmt_index_removed(int sk, void *buf, size_t len)
+{
+	struct mgmt_ev_index_removed *ev = buf;
+	uint16_t index;
+
+	if (len < sizeof(*ev)) {
+		error("Too small index removed event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	remove_controller(index);
+}
+
 static void read_mode(int sk, uint16_t index)
 {
-	char buf[MGMT_HDR_SIZE + MGMT_READ_MODE_CP_SIZE];
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_read_mode)];
 	struct mgmt_hdr *hdr = (void *) buf;
-	struct mgmt_read_mode_cp *cp = (void *) &buf[sizeof(*hdr)];
+	struct mgmt_cp_read_mode *cp = (void *) &buf[sizeof(*hdr)];
 
 	memset(buf, 0, sizeof(buf));
 	hdr->opcode = MGMT_OP_READ_MODE;
@@ -136,18 +187,18 @@ static void read_mode(int sk, uint16_t index)
 
 static void read_index_list_complete(int sk, void *buf, size_t len)
 {
-	struct mgmt_read_index_list_rp *rp = buf;
+	struct mgmt_rp_read_index_list *rp = buf;
 	uint16_t num;
 	int i;
 
-	if (len < MGMT_READ_INDEX_LIST_RP_SIZE) {
+	if (len < sizeof(*rp)) {
 		error("Too small read index list complete event");
 		return;
 	}
 
 	num = btohs(bt_get_unaligned(&rp->num_controllers));
 
-	if (num * sizeof(uint16_t) + MGMT_READ_INDEX_LIST_RP_SIZE != len) {
+	if (num * sizeof(uint16_t) + sizeof(*rp) != len) {
 		error("Incorrect packet size for index list event");
 		return;
 	}
@@ -157,8 +208,6 @@ static void read_index_list_complete(int sk, void *buf, size_t len)
 
 		index = btohs(bt_get_unaligned(&rp->index[i]));
 
-		DBG("Found controller %u", index);
-
 		add_controller(index);
 		read_info(sk, index);
 	}
@@ -166,12 +215,12 @@ static void read_index_list_complete(int sk, void *buf, size_t len)
 
 static void read_info_complete(int sk, void *buf, size_t len)
 {
-	struct mgmt_read_info_rp *rp = buf;
+	struct mgmt_rp_read_info *rp = buf;
 	struct controller_info *info;
 	uint16_t index;
 	char addr[18];
 
-	if (len < MGMT_READ_INFO_RP_SIZE) {
+	if (len < sizeof(*rp)) {
 		error("Too small read info complete event");
 		return;
 	}
@@ -192,22 +241,27 @@ static void read_info_complete(int sk, void *buf, size_t len)
 	info->type = rp->type;
 	bacpy(&info->bdaddr, &rp->bdaddr);
 	memcpy(info->features, rp->features, 8);
+	info->manufacturer = btohs(bt_get_unaligned(&rp->manufacturer));
+	info->hci_ver = rp->hci_ver;
+	info->hci_rev = btohs(bt_get_unaligned(&rp->hci_rev));
 
 	ba2str(&info->bdaddr, addr);
-	DBG("hci%u addr %s type %u", index, addr, info->type);
+	DBG("hci%u addr %s type %u manufacturer %d hci ver %d:%d",
+				index, addr, info->type, info->manufacturer,
+				info->hci_ver, info->hci_rev);
 
 	read_mode(sk, index);
 }
 
 static void read_mode_complete(int sk, void *buf, size_t len)
 {
-	struct mgmt_read_mode_rp *rp = buf;
+	struct mgmt_rp_read_mode *rp = buf;
 	struct controller_info *info;
 	uint16_t index;
 
-	if (len < MGMT_READ_MODE_RP_SIZE) {
-		error("Too small read mode complete event (%zu != %d)",
-						len, MGMT_READ_MODE_RP_SIZE);
+	if (len < sizeof(*rp)) {
+		error("Too small read mode complete event (%zu != %zu)",
+							len, sizeof(*rp));
 		return;
 	}
 
@@ -228,11 +282,16 @@ static void read_mode_complete(int sk, void *buf, size_t len)
 	info->mode = rp->mode;
 
 	DBG("hci%u enabled %u mode %u", index, info->enabled, info->mode);
+
+	manager_register_adapter(index, info->enabled);
+
+	if (info->enabled)
+		manager_start_adapter(index);
 }
 
 static void mgmt_cmd_complete(int sk, void *buf, size_t len)
 {
-	struct mgmt_cmd_complete_ev *ev = buf;
+	struct mgmt_ev_cmd_complete *ev = buf;
 	uint16_t opcode;
 
 	DBG("");
@@ -265,12 +324,32 @@ static void mgmt_cmd_complete(int sk, void *buf, size_t len)
 
 static void mgmt_cmd_status(int sk, void *buf, size_t len)
 {
-	DBG("");
+	struct mgmt_ev_cmd_status *ev = buf;
+	uint16_t opcode;
+
+	if (len < sizeof(*ev)) {
+		error("Too small management command status event packet");
+		return;
+	}
+
+	opcode = btohs(bt_get_unaligned(&ev->opcode));
+
+	DBG("status %u opcode %u", ev->status, opcode);
 }
 
 static void mgmt_controller_error(int sk, void *buf, size_t len)
 {
-	DBG("");
+	struct mgmt_ev_controller_error *ev = buf;
+	uint16_t index;
+
+	if (len < sizeof(*ev)) {
+		error("Too small management controller error event packet");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("index %u error_code %u", index, ev->error_code);
 }
 
 static gboolean mgmt_event(GIOChannel *io, GIOCondition cond, gpointer user_data)
@@ -325,6 +404,12 @@ static gboolean mgmt_event(GIOChannel *io, GIOCondition cond, gpointer user_data
 	case MGMT_EV_CONTROLLER_ERROR:
 		mgmt_controller_error(sk, buf + MGMT_HDR_SIZE, len);
 		break;
+	case MGMT_EV_INDEX_ADDED:
+		mgmt_index_added(sk, buf + MGMT_HDR_SIZE, len);
+		break;
+	case MGMT_EV_INDEX_REMOVED:
+		mgmt_index_removed(sk, buf + MGMT_HDR_SIZE, len);
+		break;
 	default:
 		error("Unknown Management opcode %u", opcode);
 		break;
@@ -478,12 +563,6 @@ static int mgmt_set_name(int index, const char *name)
 	return -ENOSYS;
 }
 
-static int mgmt_read_name(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
 static int mgmt_cancel_resolve_name(int index, bdaddr_t *bdaddr)
 {
 	char addr[18];
@@ -527,29 +606,17 @@ static int mgmt_write_eir_data(int index, uint8_t *data)
 static int mgmt_read_bdaddr(int index, bdaddr_t *bdaddr)
 {
 	char addr[18];
+	struct controller_info *info = &controllers[index];
 
-	ba2str(bdaddr, addr);
+	ba2str(&info->bdaddr, addr);
 	DBG("index %d addr %s", index, addr);
 
-	return -ENOSYS;
-}
+	if (!info->valid)
+		return -ENODEV;
 
-static int mgmt_set_event_mask(int index, uint8_t *events, size_t count)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
+	bacpy(bdaddr, &info->bdaddr);
 
-static int mgmt_write_inq_mode(int index, uint8_t mode)
-{
-	DBG("index %d mode %u", index, mode);
-	return -ENOSYS;
-}
-
-static int mgmt_read_inq_tx_pwr(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
+	return 0;
 }
 
 static int mgmt_block_device(int index, bdaddr_t *bdaddr)
@@ -580,14 +647,33 @@ static int mgmt_get_conn_list(int index, GSList **conns)
 
 static int mgmt_read_local_version(int index, struct hci_version *ver)
 {
+	struct controller_info *info = &controllers[index];
+
 	DBG("index %d", index);
-	return -ENOSYS;
+
+	if (!info->valid)
+		return -ENODEV;
+
+	memset(ver, 0, sizeof(*ver));
+	ver->manufacturer = info->manufacturer;
+	ver->hci_ver = info->hci_ver;
+	ver->hci_rev = info->hci_rev;
+
+	return 0;
 }
 
 static int mgmt_read_local_features(int index, uint8_t *features)
 {
+	struct controller_info *info = &controllers[index];
+
 	DBG("index %d", index);
-	return -ENOSYS;
+
+	if (!info->valid)
+		return -ENODEV;
+
+	memcpy(features, info->features, 8);
+
+	return 0;
 }
 
 static int mgmt_read_local_ext_features(int index)
@@ -596,12 +682,6 @@ static int mgmt_read_local_ext_features(int index)
 	return -ENOSYS;
 }
 
-static int mgmt_init_ssp_mode(int index, uint8_t *mode)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
 static int mgmt_read_link_policy(int index)
 {
 	DBG("index %d", index);
@@ -677,12 +757,6 @@ static int mgmt_read_scan_enable(int index)
 	return -ENOSYS;
 }
 
-static int mgmt_read_ssp_mode(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
 static int mgmt_write_le_host(int index, uint8_t le, uint8_t simul)
 {
 	DBG("index %d le %u simul %u", index, le, simul);
@@ -723,23 +797,18 @@ static struct btd_adapter_ops mgmt_ops = {
 	.resolve_name = mgmt_resolve_name,
 	.cancel_resolve_name = mgmt_cancel_resolve_name,
 	.set_name = mgmt_set_name,
-	.read_name = mgmt_read_name,
 	.set_class = mgmt_set_class,
 	.set_fast_connectable = mgmt_fast_connectable,
 	.read_clock = mgmt_read_clock,
 	.get_conn_handle = mgmt_conn_handle,
 	.write_eir_data = mgmt_write_eir_data,
 	.read_bdaddr = mgmt_read_bdaddr,
-	.set_event_mask = mgmt_set_event_mask,
-	.write_inq_mode = mgmt_write_inq_mode,
-	.read_inq_tx_pwr = mgmt_read_inq_tx_pwr,
 	.block_device = mgmt_block_device,
 	.unblock_device = mgmt_unblock_device,
 	.get_conn_list = mgmt_get_conn_list,
 	.read_local_version = mgmt_read_local_version,
 	.read_local_features = mgmt_read_local_features,
 	.read_local_ext_features = mgmt_read_local_ext_features,
-	.init_ssp_mode = mgmt_init_ssp_mode,
 	.read_link_policy = mgmt_read_link_policy,
 	.disconnect = mgmt_disconnect,
 	.remove_bonding = mgmt_remove_bonding,
@@ -749,7 +818,6 @@ static struct btd_adapter_ops mgmt_ops = {
 	.passkey_reply = mgmt_passkey_reply,
 	.get_auth_info = mgmt_get_auth_info,
 	.read_scan_enable = mgmt_read_scan_enable,
-	.read_ssp_mode = mgmt_read_ssp_mode,
 	.write_le_host = mgmt_write_le_host,
 	.get_remote_version = mgmt_get_remote_version,
 	.encrypt_link = mgmt_encrypt_link,
