@@ -69,11 +69,16 @@ struct remote_dev_info {
 	char *alias;
 	dbus_bool_t legacy;
 	name_status_t name_status;
+	gboolean le;
+	char **uuids;
+	size_t uuid_count;
+	GSList *services;
+	uint8_t evt_type;
+	uint8_t bdaddr_type;
+	uint8_t flags;
 };
 
 struct hci_dev {
-	int ignore;
-
 	uint8_t  features[8];
 	uint8_t  extfeatures[8];
 	uint8_t  lmp_ver;
@@ -82,14 +87,14 @@ struct hci_dev {
 	uint16_t manufacturer;
 
 	uint8_t  ssp_mode;
-	uint8_t  name[MAX_NAME_LENGTH];
+	char     name[MAX_NAME_LENGTH + 1];
 };
 
 int adapter_start(struct btd_adapter *adapter);
 
 int adapter_stop(struct btd_adapter *adapter);
 
-int adapter_update(struct btd_adapter *adapter, uint8_t cls);
+int adapter_set_service_classes(struct btd_adapter *adapter, uint8_t value);
 
 int adapter_update_ssp_mode(struct btd_adapter *adapter, uint8_t mode);
 
@@ -120,17 +125,19 @@ int adapter_get_discover_type(struct btd_adapter *adapter);
 gboolean adapter_is_ready(struct btd_adapter *adapter);
 struct remote_dev_info *adapter_search_found_devices(struct btd_adapter *adapter,
 						struct remote_dev_info *match);
+void adapter_update_device_from_info(struct btd_adapter *adapter,
+						le_advertising_info *info);
 void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 				int8_t rssi, uint32_t class, const char *name,
 				const char *alias, gboolean legacy,
 				name_status_t name_status, uint8_t *eir_data);
 int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr);
 void adapter_emit_device_found(struct btd_adapter *adapter,
-				struct remote_dev_info *dev, uint8_t *eir_data);
+				struct remote_dev_info *dev,
+				uint8_t *eir_data, size_t eir_length);
 void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode);
-void adapter_setname_complete(bdaddr_t *local, uint8_t status);
-void adapter_update_tx_power(bdaddr_t *bdaddr, uint8_t status, void *ptr);
-void adapter_update_local_name(bdaddr_t *bdaddr, uint8_t status, void *ptr);
+void adapter_update_tx_power(struct btd_adapter *adapter, int8_t tx_power);
+void adapter_update_local_name(struct btd_adapter *adapter, const char *name);
 void adapter_service_insert(const bdaddr_t *bdaddr, void *rec);
 void adapter_service_remove(const bdaddr_t *bdaddr, void *rec);
 sdp_list_t *adapter_get_services(struct btd_adapter *adapter);
@@ -177,6 +184,8 @@ int btd_adapter_restore_powered(struct btd_adapter *adapter);
 int btd_adapter_switch_online(struct btd_adapter *adapter);
 int btd_adapter_switch_offline(struct btd_adapter *adapter);
 
+typedef void (*bt_hci_result_t) (uint8_t status, gpointer user_data);
+
 struct btd_adapter_ops {
 	int (*setup) (void);
 	void (*cleanup) (void);
@@ -195,7 +204,6 @@ struct btd_adapter_ops {
 	int (*resolve_name) (int index, bdaddr_t *bdaddr);
 	int (*cancel_resolve_name) (int index, bdaddr_t *bdaddr);
 	int (*set_name) (int index, const char *name);
-	int (*read_name) (int index);
 	int (*set_class) (int index, uint32_t class);
 	int (*set_fast_connectable) (int index, gboolean enable);
 	int (*read_clock) (int index, int handle, int which, int timeout,
@@ -203,16 +211,12 @@ struct btd_adapter_ops {
 	int (*get_conn_handle) (int index, const bdaddr_t *bdaddr, int *handle);
 	int (*write_eir_data) (int index, uint8_t *data);
 	int (*read_bdaddr) (int index, bdaddr_t *bdaddr);
-	int (*set_event_mask) (int index, uint8_t *events, size_t count);
-	int (*write_inq_mode) (int index, uint8_t mode);
-	int (*read_inq_tx_pwr) (int index);
 	int (*block_device) (int index, bdaddr_t *bdaddr);
 	int (*unblock_device) (int index, bdaddr_t *bdaddr);
 	int (*get_conn_list) (int index, GSList **conns);
 	int (*read_local_version) (int index, struct hci_version *ver);
 	int (*read_local_features) (int index, uint8_t *features);
 	int (*read_local_ext_features) (int index);
-	int (*init_ssp_mode) (int index, uint8_t *ssp_mode);
 	int (*read_link_policy) (int index);
 	int (*disconnect) (int index, uint16_t handle);
 	int (*remove_bonding) (int index, bdaddr_t *bdaddr);
@@ -223,11 +227,14 @@ struct btd_adapter_ops {
 	int (*passkey_reply) (int index, bdaddr_t *bdaddr, uint32_t passkey);
 	int (*get_auth_info) (int index, bdaddr_t *bdaddr, uint8_t *auth);
 	int (*read_scan_enable) (int index);
-	int (*read_ssp_mode) (int index);
 	int (*write_le_host) (int index, uint8_t le, uint8_t simul);
+	int (*get_remote_version) (int index, uint16_t handle,
+						gboolean delayed);
+	int (*encrypt_link) (int index, bdaddr_t *bdaddr, bt_hci_result_t cb,
+							gpointer user_data);
 };
 
-int btd_register_adapter_ops(struct btd_adapter_ops *btd_adapter_ops);
+int btd_register_adapter_ops(struct btd_adapter_ops *ops, gboolean priority);
 void btd_adapter_cleanup_ops(struct btd_adapter_ops *btd_adapter_ops);
 int adapter_ops_setup(void);
 
@@ -273,7 +280,14 @@ int btd_adapter_get_auth_info(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 
 int btd_adapter_read_scan_enable(struct btd_adapter *adapter);
 
-int btd_adapter_read_ssp_mode(struct btd_adapter *adapter);
 int btd_adapter_read_local_ext_features(struct btd_adapter *adapter);
 void btd_adapter_update_local_ext_features(struct btd_adapter *adapter,
 						const uint8_t *features);
+int btd_adapter_get_remote_name(struct btd_adapter *adapter,
+							bdaddr_t *bdaddr);
+int btd_adapter_get_remote_version(struct btd_adapter *adapter,
+							uint16_t handle,
+							gboolean delayed);
+
+int btd_adapter_encrypt_link(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+				bt_hci_result_t cb, gpointer user_data);
