@@ -52,7 +52,7 @@
 #include "adapter.h"
 #include "device.h"
 #include "dbus-common.h"
-#include "dbus-hci.h"
+#include "event.h"
 #include "error.h"
 #include "glib-helper.h"
 #include "agent.h"
@@ -64,6 +64,9 @@
 #define DISCONNECT_TIMER	2
 #define DISCOVERY_TIMER		2
 
+/* When all services should trust a remote device */
+#define GLOBAL_TRUST "[all]"
+
 struct btd_driver_data {
 	guint id;
 	struct btd_device_driver *driver;
@@ -772,7 +775,6 @@ static void bonding_request_cancel(struct bonding_req *bonding)
 
 void device_request_disconnect(struct btd_device *device, DBusMessage *msg)
 {
-	GSList *l;
 	DBusConnection *conn = get_dbus_connection();
 
 	if (device->bonding)
@@ -788,21 +790,21 @@ void device_request_disconnect(struct btd_device *device, DBusMessage *msg)
 	if (device->disconn_timer)
 		return;
 
-	l = device->watches;
-	while (l) {
-		struct btd_disconnect_data *data = l->data;
-
-		l = l->next;
+	while (device->watches) {
+		struct btd_disconnect_data *data = device->watches->data;
 
 		if (data->watch)
 			/* temporary is set if device is going to be removed */
 			data->watch(device, device->temporary,
-					data->user_data);
-	}
+							data->user_data);
 
-	g_slist_foreach(device->watches, (GFunc) g_free, NULL);
-	g_slist_free(device->watches);
-	device->watches = NULL;
+		/* Check if the watch has been removed by callback function */
+		if (!g_slist_find(device->watches, data))
+			continue;
+
+		device->watches = g_slist_remove(device->watches, data);
+		g_free(data);
+	}
 
 	device->disconn_timer = g_timeout_add_seconds(DISCONNECT_TIMER,
 						do_disconnect, device);
@@ -864,7 +866,7 @@ static void device_set_connected(struct btd_device *device,
 
 		device->secmode3 = FALSE;
 
-		hcid_dbus_bonding_process_complete(&sba, &device->bdaddr, 0);
+		btd_event_bonding_process_complete(&sba, &device->bdaddr, 0);
 	}
 }
 
