@@ -301,25 +301,17 @@ void btd_event_simple_pairing_complete(bdaddr_t *local, bdaddr_t *peer,
 	device_simple_pairing_complete(device, status);
 }
 
-static char *extract_eir_name(uint8_t *data, uint8_t *type)
+void btd_event_advertising_report(bdaddr_t *local, le_advertising_info *info)
 {
-	if (!data || !type)
-		return NULL;
+	struct btd_adapter *adapter;
 
-	if (data[0] == 0)
-		return NULL;
-
-	*type = data[1];
-
-	switch (*type) {
-	case 0x08:
-	case 0x09:
-		if (!g_utf8_validate((char *) (data + 2), data[0] - 1, NULL))
-			return strdup("");
-		return strndup((char *) (data + 2), data[0] - 1);
+	adapter = manager_find_adapter(local);
+	if (adapter == NULL) {
+		error("No matching adapter found");
+		return;
 	}
 
-	return NULL;
+	adapter_update_device_from_info(adapter, info);
 }
 
 void btd_event_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
@@ -397,7 +389,7 @@ void btd_event_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 	} else
 		legacy = TRUE;
 
-	tmp_name = extract_eir_name(data, &name_type);
+	tmp_name = bt_extract_eir_name(data, &name_type);
 	if (tmp_name) {
 		if (name_type == 0x09) {
 			write_device_name(local, peer, tmp_name);
@@ -510,7 +502,7 @@ void btd_event_remote_name(bdaddr_t *local, bdaddr_t *peer, uint8_t status,
 	if (dev_info) {
 		g_free(dev_info->name);
 		dev_info->name = g_strdup(name);
-		adapter_emit_device_found(adapter, dev_info, NULL);
+		adapter_emit_device_found(adapter, dev_info, NULL, 0);
 	}
 
 	if (device)
@@ -541,12 +533,19 @@ int btd_event_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
 	if (!get_adapter_and_device(local, peer, &adapter, &device, TRUE))
 		return -ENODEV;
 
+	remote_auth = device_get_auth(device);
+
 	new_key_type = key_type;
 
 	if (key_type == 0x06) {
 		if (device_get_debug_key(device, NULL))
 			old_key_type = 0x03;
-		if (old_key_type != 0xff)
+		/* Some buggy controller combinations generate a changed
+		 * combination key for legacy pairing even when there's no
+		 * previous key */
+		if (remote_auth == 0xff && old_key_type == 0xff)
+			new_key_type = key_type = 0x00;
+		else if (old_key_type != 0xff)
 			new_key_type = old_key_type;
 		else
 			/* This is Changed Combination Link Key for
@@ -555,7 +554,6 @@ int btd_event_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
 	}
 
 	btd_adapter_get_auth_info(adapter, peer, &local_auth);
-	remote_auth = device_get_auth(device);
 	bonding = device_is_bonding(device, NULL);
 
 	DBG("key type 0x%02x old key type 0x%02x new key type 0x%02x",
@@ -715,36 +713,6 @@ void btd_event_le_set_scan_enable_complete(bdaddr_t *local, uint8_t status)
 	adapter_set_state(adapter, state);
 }
 
-void btd_event_read_simple_pairing_mode_complete(bdaddr_t *local, void *ptr)
-{
-	read_simple_pairing_mode_rp *rp = ptr;
-	struct btd_adapter *adapter;
-
-	if (rp->status)
-		return;
-
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
-		return;
-	}
-
-	adapter_update_ssp_mode(adapter, rp->mode);
-}
-
-void btd_event_write_simple_pairing_mode_complete(bdaddr_t *local)
-{
-	struct btd_adapter *adapter;
-
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
-		return;
-	}
-
-	btd_adapter_read_ssp_mode(adapter);
-}
-
 void btd_event_returned_link_key(bdaddr_t *local, bdaddr_t *peer)
 {
 	struct btd_adapter *adapter;
